#include <string.h>
#include <allegro.h>
#include <allegro/internal/aintern.h>

#include "alleggl.h"
#include "glvtable.h"
#include "allglint.h"


static BITMAP *allegro_gl_win_init_windowed(int w, int h, int v_w, int v_h,
											int color_depth);
static BITMAP *allegro_gl_win_init_fullscreen(int w, int h, int v_w, int v_h,
											  int color_depth);
static void allegro_gl_win_exit(struct BITMAP *b);

static struct allegro_gl_driver allegro_gl_win;



GFX_DRIVER gfx_allegro_gl_windowed = {
	GFX_OPENGL_WINDOWED,
	EMPTY_STRING,
	EMPTY_STRING,
	"AllegroGL Windowed (Win32)",
	allegro_gl_win_init_windowed,
	allegro_gl_win_exit,
	NULL,						/* scrolling not implemented */
	NULL,						/* vsync, may use for flip? */
	NULL,						/* No h/w pallete, not using indexed mode */
	NULL, NULL,					/* Still no scrolling */
	NULL,						/* No triple buffering */
	allegro_gl_create_video_bitmap,
	allegro_gl_destroy_video_bitmap,
	NULL, NULL,					/* No show/request video bitmaps */
	NULL, NULL,					/* No system bitmaps */
	allegro_gl_set_mouse_sprite,
	allegro_gl_show_mouse,
	allegro_gl_hide_mouse,
	allegro_gl_move_mouse,
	NULL, NULL, NULL,			/* No video state stuff */
	NULL,                       /* No fetch_mode_list */
	0, 0,						/* physical (not virtual!) screen size */
	0,							/* true if video memory is linear */
	0,							/* bank size, in bytes */
	0,							/* bank granularity, in bytes */
	0,							/* video memory size, in bytes */
	0,							/* physical address of video memory */
	TRUE                        /* Windowed mode? */
};



GFX_DRIVER gfx_allegro_gl_fullscreen = {
	GFX_OPENGL_FULLSCREEN,
	EMPTY_STRING,
	EMPTY_STRING,
	"AllegroGL Fullscreen (Win32)",
	allegro_gl_win_init_fullscreen,
	allegro_gl_win_exit,
	NULL,						/* scrolling not implemented */
	NULL,						/* vsync, may use for flip? */
	NULL,						/* No h/w pallete, not using indexed mode */
	NULL, NULL,					/* Still no scrolling */
	NULL,						/* No triple buffering */
	allegro_gl_create_video_bitmap,
	allegro_gl_destroy_video_bitmap,
	NULL, NULL,					/* No show/request video bitmaps */
	NULL, NULL,					/* No system bitmaps */
	allegro_gl_set_mouse_sprite,
	allegro_gl_show_mouse,
	allegro_gl_hide_mouse,
	allegro_gl_move_mouse,
	NULL, NULL, NULL,			/* No video state stuff */
	NULL,                       /* No fetch_mode_list */
	0, 0,						/* physical (not virtual!) screen size */
	0,							/* true if video memory is linear */
	0,							/* bank size, in bytes */
	0,							/* bank granularity, in bytes */
	0,							/* video memory size, in bytes */
	0,							/* physical address of video memory */
	FALSE                       /* Windowed mode */
};



HDC __allegro_gl_hdc = NULL;

static HGLRC allegro_glrc = NULL;
static int fullscreen = 0;
static HWND wnd = NULL;

static int initialized = 0;

static DWORD style_saved, exstyle_saved;
static DEVMODE dm_saved;
static int test_windows_created = 0;
static int new_w = 0, new_h = 0;

static PIXELFORMATDESCRIPTOR pfd = {
	sizeof(PIXELFORMATDESCRIPTOR),	/* size of this pfd */
	1,							/* version number */
	PFD_DRAW_TO_WINDOW			/* support window */
		| PFD_SUPPORT_OPENGL	/* support OpenGL */
		| PFD_DOUBLEBUFFER,		/* double buffered */
	PFD_TYPE_RGBA,				/* RGBA type */
	24,							/* 24-bit color depth */
	0, 0, 0, 0, 0, 0,			/* color bits ignored */
	0,							/* no alpha buffer */
	0,							/* shift bit ignored */
	0,							/* no accumulation buffer */
	0, 0, 0, 0,					/* accum bits ignored */
	0,							/* z-buffer */
	0,							/* no stencil buffer */
	0,							/* no auxiliary buffer */
	PFD_MAIN_PLANE,				/* main layer */
	0,							/* reserved */
	0, 0, 0						/* layer masks ignored */
};



#define ALLEGROGL_TEST_WINDOW_CLASS "AllegroGLTestWindow"
static void register_test_window()
{
	WNDCLASS wc;

	memset(&wc, 0, sizeof(wc));
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = DefWindowProc;
	wc.hInstance = GetModuleHandle(NULL);
	wc.hIcon = LoadIcon(GetModuleHandle(NULL), IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.lpszClassName = ALLEGROGL_TEST_WINDOW_CLASS;

	RegisterClass(&wc);
}



static HWND create_test_window()
{
	HWND wnd = CreateWindow(ALLEGROGL_TEST_WINDOW_CLASS,
							"AllegroGL Test Window",
							WS_POPUP | WS_CLIPCHILDREN,
							0, 0, new_w, new_h,
							NULL, NULL,
							GetModuleHandle(NULL),
							NULL);

	test_windows_created++;
	return wnd;
}



static int decode_pixel_format(PIXELFORMATDESCRIPTOR * pfd, HDC hdc, int format,
                               struct allegro_gl_display_info *dinfo,
                               int desktop_depth)
{
	TRACE("* Note * decode_pixel_format: Decoding: ");
	/* Not interested if it doesn't support OpenGL and RGBA */
	if (!(pfd->dwFlags & PFD_SUPPORT_OPENGL)) {
		TRACE("OpenGL Unsupported\n");
		return -1;
	}
	if (pfd->iPixelType != PFD_TYPE_RGBA) {
		TRACE("Not RGBA mode\n");
		return -1;
	}
	
	if ((pfd->cColorBits != desktop_depth)
	 && (pfd->cColorBits != 32 || desktop_depth < 24)) {
		TRACE("* Note * select_pixel_format: Current color depth != "
			  "pixel format color depth -> pf not tested");
		return -1;
	}
	

	/* hardware acceleration */
	if (((pfd->dwFlags & PFD_GENERIC_ACCELERATED)
		 && (pfd->dwFlags & PFD_GENERIC_FORMAT))
		|| (!(pfd->dwFlags & PFD_GENERIC_ACCELERATED)
			&& !(pfd->dwFlags & PFD_GENERIC_FORMAT)))
		dinfo->rmethod = 1;
	else
		dinfo->rmethod = 0;

	TRACE("Acceleration: %s ", ((dinfo->rmethod == 0) ? "No"
	                         : ((dinfo->rmethod == 1) ? "Yes" : "Unknown")));

	TRACE("Color Depth: %i ", pfd->cColorBits);

	/* Depths of colour buffers */
	dinfo->pixel_size.rgba.r = pfd->cRedBits;
	dinfo->pixel_size.rgba.g = pfd->cGreenBits;
	dinfo->pixel_size.rgba.b = pfd->cBlueBits;
	dinfo->pixel_size.rgba.a = pfd->cAlphaBits;
	TRACE("RGBA: %i.%i.%i.%i ", pfd->cRedBits, pfd->cGreenBits,
		  pfd->cBlueBits, pfd->cAlphaBits);

	/* Depths of accumulation buffer */
	dinfo->accum_size.rgba.r = pfd->cAccumRedBits;
	dinfo->accum_size.rgba.g = pfd->cAccumGreenBits;
	dinfo->accum_size.rgba.b = pfd->cAccumBlueBits;
	dinfo->accum_size.rgba.a = pfd->cAccumAlphaBits;
	TRACE("Accum: %i.%i.%i.%i\n\t", pfd->cAccumRedBits, pfd->cAccumGreenBits,
		  pfd->cAccumBlueBits, pfd->cAccumAlphaBits);

	/* Miscellaneous settings */
	dinfo->doublebuffered = pfd->dwFlags & PFD_DOUBLEBUFFER;
	dinfo->stereo = pfd->dwFlags & PFD_STEREO;
	dinfo->aux_buffers = pfd->cAuxBuffers;
	dinfo->depth_size = pfd->cDepthBits;
	dinfo->stencil_size = pfd->cStencilBits;
	TRACE("DblBuf: %i Zbuf: %i Stereo: %i Aux: %i Stencil: %i ",
		  dinfo->doublebuffered, dinfo->depth_size, dinfo->stereo,
		  dinfo->aux_buffers, dinfo->stencil_size);

	/* These are the component shifts, like Allegro's _rgb_*_shift_*. */
	dinfo->r_shift = pfd->cRedShift;
	dinfo->g_shift = pfd->cGreenShift;
	dinfo->b_shift = pfd->cBlueShift;
	dinfo->a_shift = pfd->cAlphaShift;
	TRACE("Shift: %i.%i.%i.%i\n\t", pfd->cRedShift, pfd->cGreenShift,
		  pfd->cBlueShift, pfd->cAlphaShift);

	dinfo->sample_buffers = 0;
	dinfo->samples = 0;

	if (allegro_gl_extensions_WGL.ARB_pixel_format
	 && allegro_gl_extensions_WGL.ARB_multisample
	 && allegro_gl_extensions_GL.ARB_multisample) {

		const int piAttributes[2] = { WGL_SAMPLE_BUFFERS_ARB,
									  WGL_SAMPLES_ARB };
		int piValues[2] = {0, 0};

		if (wglGetPixelFormatAttribivARB(hdc, format, 0, 2,
										 piAttributes, piValues)) {
			dinfo->sample_buffers = piValues[0];
			dinfo->samples = piValues[1];
		}
	}
	TRACE("Sample Buffers: %i Samples: %i\n\t",
	      dinfo->sample_buffers, dinfo->samples);

	/* This bit is the same as the X code, setting some things based on
	 * what we've read out of the PFD. */
	dinfo->colour_depth = 0;
	if (dinfo->pixel_size.rgba.r == 5 && dinfo->pixel_size.rgba.b == 5) {
		if (dinfo->pixel_size.rgba.g == 5)
			dinfo->colour_depth = 15;
		if (dinfo->pixel_size.rgba.g == 6)
			dinfo->colour_depth = 16;
	}
	if (dinfo->pixel_size.rgba.r == 8
		&& dinfo->pixel_size.rgba.g == 8 && dinfo->pixel_size.rgba.b == 8) {
		if (dinfo->pixel_size.rgba.a == 8)
			dinfo->colour_depth = 32;
		else
			dinfo->colour_depth = 24;
	}
	TRACE("Decoded bpp: %i ", dinfo->colour_depth);


	dinfo->allegro_format = (dinfo->colour_depth != 0)
		&& (dinfo->g_shift == dinfo->pixel_size.rgba.b)
		&& (dinfo->r_shift * dinfo->b_shift == 0)
		&& (dinfo->r_shift + dinfo->b_shift ==
			dinfo->pixel_size.rgba.b + dinfo->pixel_size.rgba.g);

	return 0;
}



static int decode_pixel_format_attrib(struct allegro_gl_display_info *dinfo,
                          int num_attribs, const int *attrib, const int *value,
                          int desktop_depth) {
	int i;
	
	TRACE("* Note * decode_pixel_format_attrib: Decoding: \n\t");

	for (i = 0; i < num_attribs; i++) {

		/* Not interested if it doesn't support OpenGL or window drawing or
		 * RGBA.
		 */
		if (attrib[i] == WGL_SUPPORT_OPENGL_ARB && value[i] == 0) {	
			TRACE("OpenGL Unsupported\n");
			return -1;
		}
		else if (attrib[i] == WGL_DRAW_TO_WINDOW_ARB && value[i] == 0) {	
			TRACE("Can't draw to window\n");
			return -1;
		}
		else if (attrib[i] == WGL_PIXEL_TYPE_ARB
		      && value[i] != WGL_TYPE_RGBA_ARB) {	
			TRACE("Not RGBA mode\n");
			return -1;
		}
		/* Check for color depth matching */
		else if (attrib[i] == WGL_COLOR_BITS_ARB) {
			if ((value[i] != desktop_depth)
			 && (value[i] != 32 || desktop_depth < 24)) {
				TRACE("* Note * select_pixel_format: Current color depth != "
					  "pixel format color depth -> pf not tested");
				return -1;
			}
			TRACE("Color Depth: %i ", value[i]);
		}
		/* hardware acceleration */
		else if (attrib[i] == WGL_ACCELERATION_ARB) {
			dinfo->rmethod = (value[i] == WGL_NO_ACCELERATION_ARB) ? 0 : 1;
		}
		/* Depths of colour buffers */
		else if (attrib[i] == WGL_RED_BITS_ARB) {
			dinfo->pixel_size.rgba.r = value[i];
		}
		else if (attrib[i] == WGL_GREEN_BITS_ARB) {
			dinfo->pixel_size.rgba.g = value[i];
		}
		else if (attrib[i] == WGL_BLUE_BITS_ARB) {
			dinfo->pixel_size.rgba.b = value[i];
		}
		else if (attrib[i] == WGL_ALPHA_BITS_ARB) {
			dinfo->pixel_size.rgba.a = value[i];
		}
		/* Shift of color components */
		else if (attrib[i] == WGL_RED_SHIFT_ARB) {
			dinfo->r_shift = value[i];
		}
		else if (attrib[i] == WGL_GREEN_SHIFT_ARB) {
			dinfo->g_shift = value[i];
		}
		else if (attrib[i] == WGL_BLUE_SHIFT_ARB) {
			dinfo->b_shift = value[i];
		}
		else if (attrib[i] == WGL_ALPHA_SHIFT_ARB) {
			dinfo->a_shift = value[i];
		}

		/* Depths of accumulation buffer */
		else if (attrib[i] == WGL_ACCUM_RED_BITS_ARB) {
			dinfo->accum_size.rgba.r = value[i];
		}
		else if (attrib[i] == WGL_ACCUM_GREEN_BITS_ARB) {
			dinfo->accum_size.rgba.g = value[i];
		}
		else if (attrib[i] == WGL_ACCUM_BLUE_BITS_ARB) {
			dinfo->accum_size.rgba.b = value[i];
		}
		else if (attrib[i] == WGL_ACCUM_ALPHA_BITS_ARB) {
			dinfo->accum_size.rgba.a = value[i];
		}	
		/* Miscellaneous settings */
		else if (attrib[i] == WGL_DOUBLE_BUFFER_ARB) {
			dinfo->doublebuffered = value[i];
		}
		else if (attrib[i] == WGL_STEREO_ARB) {
			dinfo->stereo = value[i];
		}
		else if (attrib[i] == WGL_AUX_BUFFERS_ARB) {
			dinfo->aux_buffers = value[i];
		}
		else if (attrib[i] == WGL_DEPTH_BITS_ARB) {
			dinfo->depth_size = value[i];
		}
		else if (attrib[i] == WGL_STENCIL_BITS_ARB) {
			dinfo->stencil_size = value[i];
		}
		/* Multisampling bits */
		else if (attrib[i] == WGL_SAMPLE_BUFFERS_ARB) {
			dinfo->sample_buffers = value[i];
		}
		else if (attrib[i] == WGL_SAMPLES_ARB) {
			dinfo->samples = value[i];
		}			
	}

	TRACE("Acceleration: %s ", ((dinfo->rmethod == 0) ? "No"
						: ((dinfo->rmethod == 1) ? "Yes" : "Unknown")));

	TRACE("RGBA: %i.%i.%i.%i ", dinfo->pixel_size.rgba.r,
	  dinfo->pixel_size.rgba.g, dinfo->pixel_size.rgba.b,
	  dinfo->pixel_size.rgba.a);
	TRACE("Accum: %i.%i.%i.%i ", dinfo->accum_size.rgba.r,
	   dinfo->accum_size.rgba.g, dinfo->accum_size.rgba.b,
	   dinfo->accum_size.rgba.a);

	TRACE("DblBuf: %i Zbuf: %i Stereo: %i Aux: %i Stencil: %i ",
		  dinfo->doublebuffered, dinfo->depth_size, dinfo->stereo,
		  dinfo->aux_buffers, dinfo->stencil_size);

	TRACE("Shift: %i.%i.%i.%i\n\t", dinfo->r_shift, dinfo->g_shift,
	                                dinfo->b_shift, dinfo->a_shift);

	TRACE("Sample Buffers: %i Samples: %i\n\t",
	      dinfo->sample_buffers, dinfo->samples);

	/* This bit is the same as the X code, setting some things based on
	 * what we've read out of the PFD. */
	dinfo->colour_depth = 0;
	if (dinfo->pixel_size.rgba.r == 5 && dinfo->pixel_size.rgba.b == 5) {
		if (dinfo->pixel_size.rgba.g == 5)
			dinfo->colour_depth = 15;
		if (dinfo->pixel_size.rgba.g == 6)
			dinfo->colour_depth = 16;
	}
	if (dinfo->pixel_size.rgba.r == 8
		&& dinfo->pixel_size.rgba.g == 8 && dinfo->pixel_size.rgba.b == 8) {
		if (dinfo->pixel_size.rgba.a == 8)
			dinfo->colour_depth = 32;
		else
			dinfo->colour_depth = 24;
	}
	TRACE("Decoded bpp: %i ", dinfo->colour_depth);

	dinfo->allegro_format = (dinfo->colour_depth != 0)
		&& (dinfo->g_shift == dinfo->pixel_size.rgba.b)
		&& (dinfo->r_shift * dinfo->b_shift == 0)
		&& (dinfo->r_shift + dinfo->b_shift ==
			dinfo->pixel_size.rgba.b + dinfo->pixel_size.rgba.g);

	return 0;
}



typedef struct format_t {
	int score;
	int format;
} format_t;



/* Helper function for sorting pixel formats */
static int select_pixel_format_sorter(const void *p0, const void *p1) {
	format_t *f0 = (format_t*)p0;
	format_t *f1 = (format_t*)p1;

	if (f0->score == f1->score) {
		return 0;
	}
	else if (f0->score > f1->score) {
		return -1;
	}
	else {
		return 1;
	}
}



void describe_pixel_format_old(HDC dc, int fmt, int desktop_depth,
                               format_t *formats, int *num_formats) {

	struct allegro_gl_display_info dinfo;
	PIXELFORMATDESCRIPTOR pfd;
	int score = -1;
	
	int result = DescribePixelFormat(dc, fmt, sizeof(pfd), &pfd);

	if (!result) {
		return;
	}
	
	result = !decode_pixel_format(&pfd, dc, fmt, &dinfo, desktop_depth);
	if (result) {
		score = __allegro_gl_score_config(fmt, &dinfo);
	}
			
	if (score < 0) {
		return; /* Reject non-compliant pixel formats */
	}

	if (formats && num_formats) {
		formats[*num_formats].score  = score;
		formats[*num_formats].format = fmt;
		(*num_formats)++;
	}

	return;
}



static AGL_GetPixelFormatAttribivARB_t __wglGetPixelFormatAttribivARB = NULL;



void describe_pixel_format_new(HDC dc, int fmt, int desktop_depth,
                               format_t *formats, int *num_formats) {
	struct allegro_gl_display_info dinfo;
	int score = -1;

	const int attrib[] = {
		WGL_SUPPORT_OPENGL_ARB,
		WGL_DRAW_TO_WINDOW_ARB,
		WGL_PIXEL_TYPE_ARB,
		WGL_ACCELERATION_ARB,
		WGL_DOUBLE_BUFFER_ARB,
		WGL_DEPTH_BITS_ARB,
		WGL_COLOR_BITS_ARB,
		WGL_RED_BITS_ARB,
		WGL_GREEN_BITS_ARB,
		WGL_BLUE_BITS_ARB,
		WGL_ALPHA_BITS_ARB,
		WGL_RED_SHIFT_ARB,
		WGL_GREEN_SHIFT_ARB,
		WGL_BLUE_SHIFT_ARB,
		WGL_ALPHA_SHIFT_ARB,
		WGL_STENCIL_BITS_ARB,
		WGL_STEREO_ARB,
		WGL_ACCUM_BITS_ARB,
		WGL_ACCUM_RED_BITS_ARB,
		WGL_ACCUM_GREEN_BITS_ARB,
		WGL_ACCUM_BLUE_BITS_ARB,
		WGL_ACCUM_ALPHA_BITS_ARB,
		WGL_AUX_BUFFERS_ARB,
		WGL_SAMPLE_BUFFERS_ARB,
		WGL_SAMPLES_ARB
	};

	const int num_attribs = sizeof(attrib) / sizeof(attrib[0]);
	int *value = (int*)malloc(sizeof(int) * num_attribs);
	int result;
	BOOL ret;

	/* Can't allocate mem? */
	if (!value) {
		return;
	}
	
	/* Get the pf attributes */
	ret = __wglGetPixelFormatAttribivARB(dc, fmt, 0, num_attribs,
	                                     attrib, value);

	if (!ret) {
		free(value);
		return;
	}

	/* Convert to AllegroGL format for scoring */
	result = !decode_pixel_format_attrib(&dinfo, num_attribs, attrib, value,
	                                     desktop_depth);
	free(value);
	
	if (result) {
		score = __allegro_gl_score_config(fmt, &dinfo);
	}

	if (score < 0) {
		return; /* Reject non-compliant pixel formats */
	}

	if (formats && num_formats) {
		formats[*num_formats].score  = score;
		formats[*num_formats].format = fmt;
		(*num_formats)++;
	}

	return;
}



/* Returns the number of pixel formats we should investigate */
int get_num_pixel_formats(HDC dc, int *new_pf_code) {
	
	/* DescribePixelFormat() returns maximum pixel format index in the old
	 * code. wglGetPixelFormatAttribivARB() does it in the new code.
	 */
	if (new_pf_code && *new_pf_code) {
		int attrib[1];
		int value[1];
		
		TRACE("* Note * get_num_pixel_formats: Attempting to use WGL_pf.\n");
		attrib[0] = WGL_NUMBER_PIXEL_FORMATS_ARB;
		if (__wglGetPixelFormatAttribivARB(dc, 0, 0, 1, attrib, value)
		                                                          == GL_FALSE) {
			TRACE("* Note * get_num_pixel_formats: WGL_pf failed!\n");
			*new_pf_code = 0;
		}
		else {
			return value[0];
		}
	}

	if (!new_pf_code || !*new_pf_code) {
		PIXELFORMATDESCRIPTOR pfd;
		TRACE("* Note * get_num_pixel_formats: Using DescribePixelFormat.\n");
		return DescribePixelFormat(dc, 1, sizeof(pfd), &pfd);
	}

	return 0;
}



/* Pick the best matching pixel format */
static int select_pixel_format(PIXELFORMATDESCRIPTOR * pfd)
{
	int i;
	int result, maxindex;
	int desktop_depth;

	HWND testwnd = NULL;
	HDC testdc   = NULL;
	HGLRC testrc = NULL;
	
	format_t *format = NULL;
	int num_formats = 0;
	int new_pf_code = 0;

	
	__allegro_gl_reset_scorer();

	/* Read again the desktop depth */
	desktop_depth = desktop_color_depth();
 
	register_test_window();
	testwnd = create_test_window();
	testdc = GetDC(testwnd);

	/* Check if we can support new pixel format code */
	{
		HDC old_dc = __allegro_gl_hdc;
		PIXELFORMATDESCRIPTOR pfd;
		int pf;
		int old_valid = __allegro_gl_valid_context;
		
		new_pf_code = 0;
		
		/* We need to create a dummy window with a pixel format to get the
		 * list of valid PFDs
		 */
		pfd.nSize = sizeof(pfd);
		pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
		pfd.iPixelType = PFD_TYPE_RGBA;

		pf = ChoosePixelFormat(testdc, &pfd);

		if (!pf) {
			TRACE("* Note * select_pixel_format: Unable to choose a PFD.\n");
			goto fail_pf;
        }

		/* Set up a GL context there */
		memset(&pfd, 0, sizeof(pfd));
		SetPixelFormat(testdc, pf, &pfd);

		testrc = wglCreateContext(testdc);
		wglMakeCurrent(testdc, testrc);

		__allegro_gl_hdc = testdc;
		 __allegro_gl_valid_context = TRUE;
		if (!allegro_gl_is_extension_supported("WGL_ARB_pixel_format")) {
			TRACE("* Note * WGL_ARB_pf unsupported.\n");
			goto fail_pf;
		}

		__wglGetPixelFormatAttribivARB = (AGL_GetPixelFormatAttribivARB_t)
		               wglGetProcAddress("wglGetPixelFormatAttribivARB");

		if (!__wglGetPixelFormatAttribivARB) {
			TRACE("* ERROR * select_pixel_format: WGL_ARB_pf not correctly "
			      "unsupported!\n");
			goto fail_pf;
		}

		new_pf_code = 1;
		goto exit_pf;

fail_pf:
		wglMakeCurrent(NULL, NULL);
		if (testrc) {
			wglDeleteContext(testrc);
		}
		testrc  = NULL;
exit_pf:
		__allegro_gl_hdc = old_dc;
		__allegro_gl_valid_context = old_valid;
	}

	maxindex = get_num_pixel_formats(testdc, &new_pf_code);

	/* Check if using the new pf code failed. Likely due to driver bug.
	 * maxindex is still valid though, so we can continue.
	 */
	if (!new_pf_code && testrc) {
		TRACE("* Warning * select_pixel_format: WGL_ARB_pf call failed - "
		      "reverted to plain old WGL.\n");
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(testrc);
		DestroyWindow(testwnd);
		testwnd = NULL;
		testrc  = NULL;
	}

	TRACE("%i formats.\n", maxindex);

	if (maxindex < 1) {
		TRACE("* ERROR * select_pixel_format: Didn't find any pixel format!\n");
		goto bail;
	}
	
	format = malloc((maxindex + 1) * sizeof(format_t));
	
	if (!format) {
		TRACE("* ERROR * select_pixel_format: Unable to allocate memory for "
		      "pixel format scores!\n");
		goto bail;
	}

	/* First, pixel formats are sorted by decreasing order */
	for (i = 1; i <= maxindex; i++) {

		TRACE("Mode %i ", i);
		if (new_pf_code) {
			describe_pixel_format_new(testdc, i, desktop_depth,
			                          format, &num_formats);
		}
		else {
			describe_pixel_format_old(testdc, i, desktop_depth,
			                          format, &num_formats);
		}
		TRACE("\n\n");
 	}

	if (new_pf_code) {
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(testrc);
		DestroyWindow(testwnd);
		testwnd = NULL;
		testrc  = NULL;
	}

	if (num_formats < 1) {
		TRACE("* ERROR * select_pixel_format: Didn't find any available pixel "
		      "format!\n");
		goto bail;
	}

	qsort(format, num_formats, sizeof(format_t), select_pixel_format_sorter);


	/* Sorted pixel formats are tested until one of them succeeds to
	 * make a GL context current */
 	for (i = 0; i < num_formats ; i++) {
		HGLRC rc;

		/* Avoid memory leaks because a window may already have been
		 * opened to get the list of pixel formats */
		if (!testwnd) {
			testwnd = create_test_window();
			testdc = GetDC(testwnd);
		}

		if (SetPixelFormat(testdc, format[i].format, pfd)) {
			rc = wglCreateContext(testdc);
			if (!rc) {
				TRACE("* Note * select_pixel_format: Unable to create RC!\n");
			}
			else {
				if (wglMakeCurrent(testdc, rc)) {
					wglMakeCurrent(testdc, NULL);
					wglDeleteContext(rc);
					rc = NULL;

					TRACE("* Note * select_pixel_format: Best config is: %i\n",
					      format[i].format);

					DescribePixelFormat(testdc, format[i].format, sizeof *pfd,
					                    pfd);

					ReleaseDC(testwnd, testdc);
					DestroyWindow(testwnd);

					result = format[i].format;
					
					free(format);
					return result;
				}
				else {
					TRACE("* Note * select_pixel_format: Couldn't make "
					      "the temporary OpenGL rendering context current for "
					      "this pixel format\n");
				}
			}
		}
		else {
			TRACE("* Note * select_pixel_format: Unable to set pixel "
			      "format!\n");
		}

		ReleaseDC(testwnd, testdc);
		DestroyWindow(testwnd);
		testwnd = NULL;
	}

	TRACE("** ERROR ** select_pixel_format: All modes have failed...\n");
bail:
	if (format) {
		free(format);
	}
	if (new_pf_code) {
		wglMakeCurrent(NULL, NULL);
		if (testrc) {
			wglDeleteContext(testrc);
		}
		if (testwnd) {
			DestroyWindow(testwnd);
		}
	}
	
 	return 0;
}



static void allegrogl_init_window(int w, int h, DWORD style, DWORD exstyle)
{
	RECT rect;

#define req __allegro_gl_required_settings
#define sug __allegro_gl_suggested_settings

	int x = 32, y = 32;
	
	if (req & AGL_WINDOW_X || sug & AGL_WINDOW_X)
		x = allegro_gl_display_info.x;
	if (req & AGL_WINDOW_Y || sug & AGL_WINDOW_Y)
		y = allegro_gl_display_info.y;

#undef req
#undef sug
	
	if (!fullscreen) {
		rect.left = x;
		rect.right = x + w;
		rect.top = y;
		rect.bottom = y + h;
	}
	else {
		rect.left = 0;
		rect.right = w;
		rect.top  = 0;
		rect.bottom = h;
	}

	/* save original Allegro styles */
	style_saved = GetWindowLong(wnd, GWL_STYLE);
	exstyle_saved = GetWindowLong(wnd, GWL_EXSTYLE);

	/* set custom AllegroGL style */
	SetWindowLong(wnd, GWL_STYLE, style);
	SetWindowLong(wnd, GWL_EXSTYLE, exstyle);

	if (!fullscreen) {
		AdjustWindowRectEx(&rect, style, FALSE, exstyle);
	}

	/* make the changes visible */
	SetWindowPos(wnd, 0, rect.left, rect.top,
		rect.right - rect.left, rect.bottom - rect.top,
		SWP_NOZORDER | SWP_FRAMECHANGED);
	
	return;
}



static BITMAP *allegro_gl_create_screen (GFX_DRIVER *drv, int w, int h,
                                         int depth)
{
	BITMAP *bmp;
	int is_linear = drv->linear;

	drv->linear = 1;
	bmp = _make_bitmap (w, h, 0, drv, depth, 0);
	
	if (!bmp) {
		return NULL;
	}

	bmp->id = BMP_ID_VIDEO | 1000;
	drv->linear = is_linear;

	drv->w = w;
	drv->h = h;

	return bmp;
}



static BITMAP *allegro_gl_win_init(int w, int h, int v_w, int v_h)
{
	static int first_time = 1;
	
	DWORD style=0, exstyle=0;
	int desktop_depth;
	int pf=0;

	new_w = w;
	new_h = h;

	/* virtual screen are not supported */
	if ((v_w != 0 && v_w != w) || (v_h != 0 && v_h != h))
		return NULL;
		
	/* Fill in missing color depth info */
	__allegro_gl_fill_in_info();

	/* Be sure the current desktop color depth is at least 15bpp */
	/* We may want to change this, so try to set a better depth, or
	   to at least report an error somehow */
	desktop_depth = desktop_color_depth();

	if (desktop_depth < 15)
		return NULL;

	TRACE("* Note * win_init: Requested color depth: %i  "
	      "Desktop color depth: %i\n", allegro_gl_display_info.colour_depth,
	      desktop_depth);

	/* XXX <rohannessian> I don't recall why we need to set a NULL window
	 * here, but if we do, it messes up the fullscreen mode input (code as
	 * is), or both FS and windowed mode inputs (always calling
	 * win_set_window()).
	 *
	 * I do believe this code should be commented out.
	 */
	/* request a fresh new window from Allegro... */
	/*
	if (!first_time) {
		win_set_window(NULL);
	}
	*/
	first_time = 0;

	/* ...and retrieve its handle */
	wnd = win_get_window();
	if (!wnd)
		return NULL;
	
	TRACE("* Note * win_init: Driver selected fullscreen: %s\n",
	      fullscreen ? "Yes" : "No");

	if (fullscreen)
	{
		DEVMODE dm;
    	
		int bpp_to_check[] = {16, 32, 24, 15, 0};
		int bpp_checked[] = {0, 0, 0, 0, 0};
		int bpp_index = 0;
		int i, j, result, modeswitch, done = 0;

		gfx_allegro_gl_fullscreen.w = w;
		gfx_allegro_gl_fullscreen.h = h;
		
		for (j = 0; j < 4; j++)
		{
			if (bpp_to_check[j] == allegro_gl_get(AGL_COLOR_DEPTH))
			{
				bpp_index = j;
				break;
			}
		}

		dm.dmSize = sizeof(DEVMODE);
		dm_saved.dmSize = sizeof(DEVMODE);
		
		/* Save old mode */
		EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &dm_saved);
		dm.dmBitsPerPel = desktop_depth; /* Go around Win95's bug */

		do
		{
			if (!bpp_to_check[bpp_index])
			{
				TRACE("** ERROR ** win_init: No more color depths to test.\n");
				TRACE("\tUnable to find appropriate full screen mode and pixel "
				      "format.\n");
				goto Error;
			}

			TRACE("* Note * win_init: Testing color depth: %i\n",
			      bpp_to_check[bpp_index]);
			
			memset(&dm, 0, sizeof(DEVMODE));
			dm.dmSize = sizeof(DEVMODE);
            
			i = 0;
			do 
			{
				modeswitch = EnumDisplaySettings(NULL, i, &dm);
				if (!modeswitch)
					break;
				i++;
			}
			while ((dm.dmPelsWidth != (unsigned) w) || (dm.dmPelsHeight != (unsigned) h) || (dm.dmBitsPerPel != (unsigned) bpp_to_check[bpp_index]));

			if (!modeswitch) {
				TRACE("* Note * win_init: Unable to set mode, continuing with "
				      "next color depth\n");
			}
			else
			{
				TRACE("* Note * win_init: bpp_to_check[bpp_index] = %i\n",
				      bpp_to_check[bpp_index]);
				TRACE("* Note * win_init: dm.dmBitsPerPel = %i\n",
				      (int)dm.dmBitsPerPel);

				dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;

				result = ChangeDisplaySettings(&dm, CDS_FULLSCREEN);

				if (result == DISP_CHANGE_SUCCESSFUL) 
				{
					pf = select_pixel_format(&pfd);
					if (pf)
					{
						TRACE("mode found\n");
						done = 1;
					}
				}

				style = WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
				exstyle = WS_EX_APPWINDOW | WS_EX_TOPMOST;
			}

			bpp_checked[bpp_index] = 1;

			bpp_index = 0;
			while (bpp_checked[bpp_index]) {
				bpp_index++;
			}
		} while (!done);
	}
	else
	{
		gfx_allegro_gl_windowed.w = w;
		gfx_allegro_gl_windowed.h = h;
		style = WS_SYSMENU | WS_CAPTION | WS_MINIMIZEBOX | WS_CLIPCHILDREN
		      | WS_CLIPSIBLINGS;
		exstyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
	}

	/* set up the AllegroGL window */
	allegrogl_init_window(w, h, style, exstyle);

	__allegro_gl_hdc = GetDC(wnd); /* get the device context of our window */
	if (!__allegro_gl_hdc) {
		goto Error;
	}


	if (!fullscreen)
	{
		pf = select_pixel_format(&pfd);
		if (pf == 0)
			goto Error;
	}

	/* set the pixel format */
	if (!SetPixelFormat(__allegro_gl_hdc, pf, &pfd)) { 
		TRACE("** ERROR ** win_init: Unable to set pixel format.\n");
		goto Error;
	}

	decode_pixel_format(&pfd, __allegro_gl_hdc, pf, &allegro_gl_display_info, desktop_depth);
	__allegro_gl_set_allegro_image_format(FALSE);
	set_color_depth(allegro_gl_display_info.colour_depth);
	allegro_gl_display_info.w = w;
	allegro_gl_display_info.h = h;

	/* create an OpenGL context */
	allegro_glrc = wglCreateContext(__allegro_gl_hdc);
	
	if (!allegro_glrc) { /* make the context the current one */
		goto Error;
	}
	if (!wglMakeCurrent(__allegro_gl_hdc, allegro_glrc)) {
		goto Error;
	}

	/* XXX <rohannessian> Win98/2k/XP's window forground rules don't let us
	 * make the window the topmost window on launch. This causes issues on 
	 * full-screen apps. So we use this trick to force the window to be topmost.
	 * 
	 * See http://support.microsoft.com:80/support/kb/articles/Q97/9/25.asp
	 * for details.
	 */
	{
		DWORD lock_time;

#define SPI_GETFOREGROUNDLOCKTIMEOUT 0x2000
#define SPI_SETFOREGROUNDLOCKTIMEOUT 0x2001
		if (fullscreen) {
			SystemParametersInfo(SPI_GETFOREGROUNDLOCKTIMEOUT,
			                     0, (LPVOID)&lock_time, 0);
			SystemParametersInfo(SPI_SETFOREGROUNDLOCKTIMEOUT,
			                     0, (LPVOID)0,
			                     SPIF_SENDWININICHANGE | SPIF_UPDATEINIFILE);
		}

		ShowWindow(wnd, SW_SHOWNORMAL);
		SetForegroundWindow(wnd);
		UpdateWindow(wnd);

		if (fullscreen) {
			SystemParametersInfo(SPI_SETFOREGROUNDLOCKTIMEOUT,
			                     0, (LPVOID)lock_time,
			                     SPIF_SENDWININICHANGE | SPIF_UPDATEINIFILE);
		}
#undef SPI_GETFOREGROUNDLOCKTIMEOUT
#undef SPI_SETFOREGROUNDLOCKTIMEOUT
	}
		
	win_grab_input();
	
	if (fullscreen)	{
		allegro_gl_screen= allegro_gl_create_screen(&gfx_allegro_gl_fullscreen,
		                                w, h, allegro_gl_get(AGL_COLOR_DEPTH));
	}
	else {
		allegro_gl_screen= allegro_gl_create_screen(&gfx_allegro_gl_windowed,
		                                w, h, allegro_gl_get(AGL_COLOR_DEPTH));
	}

	if (!allegro_gl_screen) {
		ChangeDisplaySettings(NULL, 0);
		goto Error;
	}
	
	TRACE("* Note * win_init: GLScreen: %ix%ix%i\n",
	      w, h, allegro_gl_get(AGL_COLOR_DEPTH));

	allegro_gl_screen->id |= BMP_ID_VIDEO | 1000;

	__allegro_gl_valid_context = TRUE;
	__allegro_gl_driver = &allegro_gl_win;
	initialized = 1;

	/* Print out OpenGL version info */
	TRACE("\nOpenGL Version: %s\nVendor: %s\nRenderer: %s\n\n",
	      (AL_CONST char*)glGetString(GL_VERSION),
	      (AL_CONST char*)glGetString(GL_VENDOR),
	      (AL_CONST char*)glGetString(GL_RENDERER));

	/* Detect if the GL driver is based on Mesa */
	allegro_gl_info.is_mesa_driver = FALSE;
	if (strstr((AL_CONST char*)glGetString(GL_VERSION),"Mesa")) {
		AGL_LOG(1, "OpenGL driver based on Mesa\n");
		allegro_gl_info.is_mesa_driver = TRUE;
	}

	/* init the GL extensions */
	__allegro_gl_manage_extensions();
	
	/* Update screen vtable in order to use AGL's */
	__allegro_gl__glvtable_update_vtable(&allegro_gl_screen->vtable);
	__allegro_gl_init_screen_mode();
	memcpy(&_screen_vtable, allegro_gl_screen->vtable, sizeof(GFX_VTABLE));
	allegro_gl_screen->vtable = &_screen_vtable;

	/* Print out WGL extension info */
	if (wglGetExtensionsStringARB) {
		AGL_LOG(1, "WGL Extensions :\n");
#if LOGLEVEL >= 1
		__allegro_gl_print_extensions((AL_CONST char*)wglGetExtensionsStringARB(wglGetCurrentDC()));
#endif
	}
	else {
		TRACE("* Note * win_init: No WGL Extensions available\n");
	}

	gfx_capabilities |= GFX_HW_CURSOR;

	/* Initialize a reasonable viewport. Those should be OpenGL defaults,
	 * but some drivers don't implement this correctly.
	 */	
	glViewport(0, 0, SCREEN_W, SCREEN_H);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	/* Set up some variables that some GL drivers omit */
	glBindTexture(GL_TEXTURE_2D, 0);
	
	return allegro_gl_screen;
	
Error:
	if (allegro_glrc) {
		wglDeleteContext(allegro_glrc);
	}
	if (__allegro_gl_hdc) {
		ReleaseDC(wnd, __allegro_gl_hdc);
	}
	__allegro_gl_hdc = NULL;
	ChangeDisplaySettings(NULL, 0);
	allegro_gl_win_exit(NULL);

	return NULL;
}



static BITMAP *allegro_gl_win_init_windowed(int w, int h, int v_w, int v_h,
											int color_depth)
{
	fullscreen = 0;
	return allegro_gl_win_init(w, h, v_w, v_h);
}



static BITMAP *allegro_gl_win_init_fullscreen(int w, int h, int v_w, int v_h,
											  int color_depth)
{
	fullscreen = 1;
	return allegro_gl_win_init(w, h, v_w, v_h);
}



static void allegro_gl_win_exit(struct BITMAP *b)
{
	__allegro_gl_unmanage_extensions();
	
	if (allegro_glrc) {
		wglDeleteContext(allegro_glrc);
		allegro_glrc = NULL;
	}
		
	if (__allegro_gl_hdc) {
		ReleaseDC(wnd, __allegro_gl_hdc);
		__allegro_gl_hdc = NULL;
	}

	if (fullscreen && initialized) {
		/* Restore screen */
		ChangeDisplaySettings(NULL, 0);
	}
	initialized = 0;

	/* hide the window */
	system_driver->restore_console_state();

	/* restore original Allegro styles */
	SetWindowLong(wnd, GWL_STYLE, style_saved);
	SetWindowLong(wnd, GWL_EXSTYLE, exstyle_saved);
	SetWindowPos(wnd, 0, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER
	                               | SWP_FRAMECHANGED);

	__allegro_gl_release_screen();
	__allegro_gl_valid_context = FALSE;
	
	return;
}



/* AllegroGL driver routines */

static void flip(void)
{
	SwapBuffers(__allegro_gl_hdc);
}



static void gl_on(void)
{
	return;
}



static void gl_off(void)
{
	return;
}



/* AllegroGL driver */

static struct allegro_gl_driver allegro_gl_win = {
	flip, gl_on, gl_off, NULL
};

